@@ -2,8 +2,7 @@
  * files.h -- header file for files.c
  * Direct file manipulation for irc?  Unheard of!
  *
- * Copyright 1995 Jeremy Nelson
- * Copyright 1997 EPIC Software Labs
+ * Copyright 1995, 1997 EPIC Software Labs
  * See the COPYRIGHT file for copyright information
  */
 
@@ -23,8 +22,9 @@
 	int	file_valid		(int);
 	int	file_rewind		(int);
 	int	file_error		(int);
-	int	file_seek		(int, long, const char *);
-	int	file_tell		(int);
+	int	file_seek		(int, off_t, const char *);
+	intmax_t file_tell		(int);
 	int	file_skip		(int, int);
 
+	char *	dbmctl			(char *);
 #endif
