@@ -12,14 +12,12 @@
 typedef struct {
 	const char	*command;
 	void 		(*inbound_handler) (const char *, const char *, const char **);
-	void		(*outbound_handler) (char *);
 	int		flags;
 } protocol_command;
 extern 	protocol_command rfc1459[];
 extern	int		 num_protocol_cmds;
 
-#define PROTO_NOQUOTE 	1 << 0
-#define PROTO_DEPREC	1 << 1
+#define PROTO_QUOTEBAD 	1 << 0
 
 
 	void    rfc1459_odd 	(const char *, const char *, const char **);
@@ -28,9 +26,6 @@ const 	char	*PasteArgs 	(const char **, int);
 	int	is_channel	(const char *);
 
 extern	const char	*FromUserHost;
-#if 0
-extern	int	doing_privmsg;
-#endif
 
 #endif
 
