@@ -1,4 +1,4 @@
-/* $EPIC: array.c,v 1.17 2003/10/31 08:19:24 crazyed Exp $ */
+/* $EPIC: array.c,v 1.20 2005/10/04 01:25:50 jnelson Exp $ */
 /*
  * array.c -- Karll's Array Suite
  *
@@ -200,6 +200,7 @@ Karll on IRC
 #include "output.h"
 #include "functions.h"
 #include "words.h"
+#include "reg.h"
 #undef BUILT_IN_FUNCTION
 #define BUILT_IN_FUNCTION(x, y) char * x (char * y)
 #undef index			/* doh! */
@@ -228,14 +229,14 @@ static int compare_indices (const void *a1, const void *a2)
 {
 	int result;
 
-	result = strcmp(qsort_array->item[*(const int*)a1],
-			qsort_array->item[*(const int*)a2]);
+	result = strcmp(qsort_array->item[*(const long*)a1],
+			qsort_array->item[*(const long*)a2]);
 
 	/* array is (to be) sorted by name, then by item number. */
 	if (result)
 		return result;
 	else
-		return *(const int*)a1 - *(const int*)a2;
+		return *(const long*)a1 - *(const long*)a2;
 }
 
 static void sort_indices (an_array *array)
@@ -786,12 +787,8 @@ BUILT_IN_FUNCTION(function_delitem, input)
 			item = my_atol(itemstr);
 			if (item >= 0 && item < array->size)
 			{
-				found = 0;
 				if (array->size == 1)
-				{
 					delete_array(name);
-					break;
-				}
 				else
 				{
 					if (item == array->index[item])
@@ -813,6 +810,7 @@ BUILT_IN_FUNCTION(function_delitem, input)
 					RESIZE(array->item, char *, array->size);
 					RESIZE(array->index, long, array->size);
 				}
+				found = 0;
 			}
 		}
 	}
@@ -842,7 +840,6 @@ BUILT_IN_FUNCTION(function_delitems, input)
 			item = my_atol(itemstr);
 			if (item >= 0 && item < array->size && array->item[item])
 			{
-				found = 0;
 				deleted++;
 				if (deleted >= array->size)
 				{
@@ -851,6 +848,7 @@ BUILT_IN_FUNCTION(function_delitems, input)
 					break;
 				}
 				new_free(&array->item[item]);
+				found = 0;
 			}
 		}
 		if (deleted)
@@ -927,7 +925,7 @@ BUILT_IN_FUNCTION(function_listarray, input)
 
 	if ((name = next_arg(input, &input)) && (array = get_array(name)))
 	{
-		char *separator = (input && *input) ? new_next_arg(input, &input) : space;
+		const char *separator = (input && *input) ? new_next_arg(input, &input) : space;
 
 		for (idx = 0; idx < array->size; idx++)
 			malloc_strcat_wordlist_c(&result, separator, array->item[idx], &resclue);
