@@ -1,4 +1,4 @@
-/* $EPIC: compat.c,v 1.24 2008/03/17 03:42:46 jnelson Exp $ */
+/* $EPIC: compat.c,v 1.36 2009/06/18 04:42:58 jnelson Exp $ */
 /*
  * Everything that im not directly responsible for I put in here.  Almost
  * all of this stuff is either borrowed from somewhere else (for you poor
@@ -979,6 +979,7 @@ size_t	strlcat (char *dst, const char *src, size_t siz)
 #endif
 
 /* --------------------------- start of arc4 stuff -------------------- */
+#ifndef HAVE_ARC4RANDOM
 /*
  * Arc4 random number generator for OpenBSD.
  * Copyright 1996 David Mazieres <dm@lcs.mit.edu>.
@@ -1004,9 +1005,9 @@ size_t	strlcat (char *dst, const char *src, size_t siz)
  * RC4 is a registered trademark of RSA Laboratories.
  */
 struct bsd_arc4_stream {
-	u_char	i;
-	u_char	j;
-	u_char	s[256];
+	unsigned char	i;
+	unsigned char	j;
+	unsigned char	s[256];
 };
 typedef struct bsd_arc4_stream 	ARC4;
 
@@ -1025,10 +1026,10 @@ static void	bsd_arc4_init (ARC4 *as)
 }
 
 __inline__
-static void	bsd_arc4_addrandom (ARC4 *as, u_char *dat, int datlen)
+static void	bsd_arc4_addrandom (ARC4 *as, unsigned char *dat, int datlen)
 {
 	int     n;
-	u_char	si;
+	unsigned char	si;
 
 	as->i--;
 	for (n = 0; n < 256; n++) {
@@ -1046,7 +1047,7 @@ static void	bsd_arc4_stir (ARC4 *as)
 	struct {
 		Timeval tv;
 		pid_t 	pid;
-		u_char	rnd[128 - sizeof(Timeval) - sizeof(pid_t)];
+		unsigned char	rnd[128 - sizeof(Timeval) - sizeof(pid_t)];
 	}       rdat;
 
 	gettimeofday(&rdat.tv, NULL);
@@ -1063,9 +1064,9 @@ static void	bsd_arc4_stir (ARC4 *as)
 }
 
 __inline__
-static u_char		bsd_arc4_getbyte (ARC4 *as)
+static unsigned char		bsd_arc4_getbyte (ARC4 *as)
 {
-	u_char si, sj;
+	unsigned char si, sj;
 
 	as->i = (as->i + 1);
 	si = as->s[as->i];
@@ -1097,7 +1098,7 @@ void	bsd_arc4random_stir (void)
 	bsd_arc4_stir(&rs);
 }
 
-void	bsd_arc4random_addrandom (u_char *dat, int datlen)
+void	bsd_arc4random_addrandom (unsigned char *dat, int datlen)
 {
 	if (!rs_initialized)
 		bsd_arc4random_stir();
@@ -1110,9 +1111,131 @@ u_32int_t	bsd_arc4random (void)
 		bsd_arc4random_stir();
 	return bsd_arc4_getword(&rs);
 }
+#endif
+
+#if 0
+/* ----------------------- start of humanize_number stuff ------------------*/
+/*	$NetBSD: humanize_number.c,v 1.8 2004/07/27 01:56:24 enami Exp $ */ 
+/*
+ * Copyright (c) 1997, 1998, 1999, 2002 The NetBSD Foundation, Inc.
+ * All rights reserved.
+ *
+ * This code is derived from software contributed to The NetBSD Foundation
+ * by Jason R. Thorpe of the Numerical Aerospace Simulation Facility,
+ * NASA Ames Research Center, by Luke Mewburn and by Tomas Svensson.
+ *
+ * This is licensed under the 4-clause BSD license, which is the 3-clause
+ * BSD license (found above) with this clause added:
+ * 3. All advertising materials mentioning features or use of this software
+ *    must display the following acknowledgement:
+ *      This product includes software developed by the NetBSD
+ *      Foundation, Inc. and its contributors.
+ */
+
+/* XXX TODO -- At some point 'bytes' needs to be 64 bits */
+int humanize_number (char *buf, size_t len, long bytes, const char *suffix, int scale, int flags)
+{
+	const char *prefixes, *sep;
+	int	b, i, r, maxscale, s1, s2, sign;
+	/* XXX TODO -- divisor and max need to become 64 bits */
+	long	divisor, max;
+	size_t	baselen;
+
+	if (buf == NULL)
+		return -1;
+	if (suffix == NULL)
+		return -1;
+	if (scale < 0)
+		return -1;
+
+	if (flags & HN_DIVISOR_1000) {
+		/* SI for decimal multiplies */
+		divisor = 1000;
+		if (flags & HN_B)
+			prefixes = "B\0k\0M\0G\0T\0P\0E";
+		else
+			prefixes = "\0\0k\0M\0G\0T\0P\0E";
+	} else {
+		/*
+		 * binary multiplies
+		 * XXX IEC 60027-2 recommends Ki, Mi, Gi...
+		 */
+		divisor = 1024;
+		if (flags & HN_B)
+			prefixes = "B\0K\0M\0G\0T\0P\0E";
+		else
+			prefixes = "\0\0K\0M\0G\0T\0P\0E";
+	}
+
+#define	SCALE2PREFIX(scale)	(&prefixes[(scale) << 1])
+	maxscale = 7;
+
+	if (scale >= maxscale &&
+	    (scale & (HN_AUTOSCALE | HN_GETSCALE)) == 0)
+		return (-1);
+
+	if (buf == NULL || suffix == NULL)
+		return (-1);
+
+	if (len > 0)
+		buf[0] = '\0';
+	if (bytes < 0) {
+		sign = -1;
+		bytes *= -100;
+		baselen = 3;		/* sign, digit, prefix */
+	} else {
+		sign = 1;
+		bytes *= 100;
+		baselen = 2;		/* digit, prefix */
+	}
+	if (flags & HN_NOSPACE)
+		sep = "";
+	else {
+		sep = " ";
+		baselen++;
+	}
+	baselen += strlen(suffix);
+
+	/* Check if enough room for `x y' + suffix + `\0' */
+	if (len < baselen + 1)
+		return (-1);
+
+	if (scale & (HN_AUTOSCALE | HN_GETSCALE)) {
+		/* See if there is additional columns can be used. */
+		for (max = 100, i = len - baselen; i-- > 0;)
+			max *= 10;
+
+		for (i = 0; bytes >= max && i < maxscale; i++)
+			bytes /= divisor;
+
+		if (scale & HN_GETSCALE)
+			return (i);
+	} else
+		for (i = 0; i < scale && i < maxscale; i++)
+			bytes /= divisor;
+
+	/* If a value <= 9.9 after rounding and ... */
+	if (bytes < 995 && i > 0 && flags & HN_DECIMAL) {
+		/* baselen + \0 + .N */
+		if (len < baselen + 1 + 2)
+			return (-1);
+		b = ((int)bytes + 5) / 10;
+		s1 = b / 10;
+		s2 = b % 10;
+		r = snprintf(buf, len, "%d.%d%s%s%s",
+		    sign * s1, s2, sep, SCALE2PREFIX(i), suffix);
+	} else
+		r = snprintf(buf, len, "%lld%s%s%s",
+		    /* LONGLONG */
+		    (long long)(sign * ((bytes + 50) / 100)),
+		    sep, SCALE2PREFIX(i), suffix);
+
+	return (r);
+}
+#endif
 
 /* --------------------------- start of misc stuff -------------------- */
-/* This is all written by Jeremy Nelson and is public domain */
+/* This is written by EPIC Software Labs contributers and is public domain */
 #ifndef HAVE_VSNPRINTF
 int vsnprintf (char *str, size_t size, const char *format, va_list ap)
 {
@@ -1121,10 +1244,10 @@ int vsnprintf (char *str, size_t size, const char *format, va_list ap)
 	/* If the string ended up overflowing, just give up. */
 	/* Pre-ansi vsprintf()s return (char *) */
 	if (ret == (int)str && strlen(str) > size)
-		panic("Buffer overflow in vsnprintf");
+		panic(1, "Buffer overflow in vsnprintf");
 	/* ANSI vsprintf()s return (int) */
 	if (ret != (int)str && ret > size)
-		panic("Buffer overflow in vsnprintf");
+		panic(1, "Buffer overflow in vsnprintf");
 
 	/* We always return (int). */
 	return ret;
@@ -1216,38 +1339,22 @@ char *	my_realpath (const char *pathname, char resolved_path[MAXPATHLEN])
 }
 #endif
 
-#ifndef HAVE_ATOLL
-# ifdef HAVE_LONG_LONG
-#  ifdef HAVE_STRTOLL
-#   define HAVE_ATOLL_REPLACEMENT
-long long	atoll (const char *str)
+/*
+#ifdef NEED_STRTOLL
+long long	strtoll (const char *nptr, char **endptr, int base)
 {
-	return strtoll(str, NULL, 0);
 }
-#  else
-#   ifdef HAVE_ATOQ
-#    define HAVE_ATOLL_REPLACEMENT
-long long	atoll (const char *str)
-{
-	return (long long)atoq(str);
-}
-#   endif
-#  endif
-# endif
-# ifndef HAVE_ATOLL_REPLACEMENT
-#  ifdef HAVE_LONG_LONG
-long long	atoll (const char *str)
-{
-	return (long long)atol(str);
-}
-#  else
-long	atoll (const char *str)
-{
-	return atol(str);
-}
-#  endif
-# endif
 #endif
+*/
+
+/*
+#ifndef HAVE_STRTOIMAX
+long		strtoimax (const char *nptr, char **endptr, int base)
+{
+	return (long)strtol(nptr, endptr, base);
+}
+#endif
+*/
 
 /**** END MISC PUBLIC DOMAIN STUFF ****/
 
