@@ -1,6 +1,6 @@
-/* $EPIC: glob.c,v 1.10 2003/07/22 21:12:54 jnelson Exp $ */
+/* $EPIC: glob.c,v 1.13 2008/04/04 04:51:05 jnelson Exp $ */
 #include "config.h"
-#if defined(NEED_GLOB)
+#ifdef NEED_GLOB
 
 /*
  * Copyright (c) 1989, 1993
@@ -63,7 +63,9 @@
  *	Number of matches in the current invocation of glob.
  */
 
+#ifdef HAVE_SYS_PARAM_H
 #include <sys/param.h>
+#endif
 #include <sys/stat.h>
 #include <ctype.h>
 #include <dirent.h>
@@ -75,6 +77,7 @@
 #include <string.h>
 #include <unistd.h>
 #include "irc.h"
+#include "compat.h"
 #ifndef MAXPATHLEN
 # ifndef PATHSIZE
 #  define MAXPATHLEN 1024
@@ -82,7 +85,6 @@
 #  define MAXPATHLEN PATHSIZE
 # endif
 #endif
-#include "compat.h"
 
 #undef EOS
 #define	DOLLAR		'$'
@@ -146,11 +148,11 @@ int bsd_glob		(	const char *pattern,
 				int (*errfunc) (const char *, int),
 				glob_t *pglob				)
 {
-	const u_char *patnext;
+	const unsigned char *patnext;
 	int c;
 	Char *bufnext, *bufend, patbuf[MAXPATHLEN+1];
 
-	patnext = (const u_char *) pattern;
+	patnext = (const unsigned char *) pattern;
 	if (!(flags & GLOB_APPEND)) 
 	{
 		pglob->gl_pathc = 0;
@@ -621,14 +623,14 @@ static int glob3		(	Char *pathbuf,
 
 	while ((dp = (*readdirfunc)(dirp))) 
 	{
-		register u_char *sc;
+		register unsigned char *sc;
 		register Char *dc;
 		int	nocase = 0;
 
 		/* Initial DOT must be matched literally. */
 		if (dp->d_name[0] == DOT && *pattern != DOT)
 			continue;
-		for (sc = (u_char *) dp->d_name, dc = pathend; 
+		for (sc = (unsigned char *) dp->d_name, dc = pathend; 
 		     (*dc++ = *sc++) != EOS;)
 			continue;
 
