@@ -1,4 +1,4 @@
-/* $EPIC: lastlog.c,v 1.29 2003/11/24 23:32:24 jnelson Exp $ */
+/* $EPIC: lastlog.c,v 1.81 2009/09/11 21:02:02 jnelson Exp $ */
 /*
  * lastlog.c: handles the lastlog features of irc. 
  *
@@ -33,8 +33,8 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  */
-
 #include "irc.h"
+#include "levels.h"
 #include "lastlog.h"
 #include "window.h"
 #include "screen.h"
@@ -43,196 +43,225 @@
 #include "output.h"
 #include "numbers.h"
 #include "functions.h"
-#include <regex.h>
+#include "reg.h"
+#include "alias.h"
 
-static int	show_lastlog (Lastlog **l, int *skip, int *number, int, char *match, regex_t *reg, int *max, const char *target);
+typedef struct	lastlog_stru
+{
+	int	level;
+	char	*target;
+	char	*msg;
+	struct	lastlog_stru	*older;
+	struct	lastlog_stru	*newer;
+	time_t	when;
+	int	visible;
+	intmax_t refnum;
+	unsigned winref;
+	int	dead;
+}	Lastlog;
 
+static	intmax_t global_lastlog_refnum = 0;
+
+static int	show_lastlog (Lastlog **l, int *skip, int *number, Mask *level_mask, char *match, regex_t *rex, int *max, const char *target, int mangler, unsigned winref, char **);
+static int	oldest_lastlog_for_window (Lastlog **item, unsigned winref);
+static int	newer_lastlog_entry (Lastlog **item, unsigned winref);
+static int	older_lastlog_entry (Lastlog **item, unsigned winref);
+static int	newest_lastlog_for_window (Lastlog **item, unsigned winref);
+static void	remove_lastlog_item (Lastlog *item);
+static void	switch_lastlog_window (Lastlog *item, unsigned newref);
+static void	move_lastlog_item (Lastlog *item, unsigned newref);
+
+Lastlog *	lastlog_oldest = NULL;
+Lastlog *	lastlog_newest = NULL;
+
+/**********************************************************************/
 /*
  * lastlog_level: current bitmap setting of which things should be stored in
  * the lastlog.  The LOG_MSG, LOG_NOTICE, etc., defines tell more about this 
  */
-static	int	lastlog_level;
-static	int	notify_level;
-	int	new_server_lastlog_level;
-	int	beep_on_level;
-	int 	current_window_level = 0;
+static	Mask	lastlog_mask;
+static	Mask	notify_mask;
+	Mask *	new_server_lastlog_mask = NULL;
+	Mask *	old_server_lastlog_mask = NULL;
+	Mask 	current_window_mask;
 
 /*
- * msg_level: the mask for the current message level.  What?  Did he really
- * say that?  This is set in the set_lastlog_msg_level() routine as it
- * compared to the lastlog_level variable to see if what ever is being added
- * should actually be added 
- */
-static	int	msg_level = LOG_CRAP;
-
-#define NUMBER_OF_LEVELS 16
-static	const char	*levels[] =
-{
-	"CRAP",		"PUBLIC",	"MSGS",		"NOTICES",
-	"WALLS",	"WALLOPS",	"NOTES",	"OPNOTES",
-	"SNOTES",	"ACTIONS",	"DCC",		"CTCP",
-	"USERLOG1",	"USERLOG2",	"USERLOG3",	"USERLOG4"
-};
-
-/*
- * warn_lastlog_level: tells the user what levels are available.
- */
-static void	warn_lastlog_levels (void)
-{
-	char buffer[BIG_BUFFER_SIZE + 1];
-	int i;
-
-	strlcpy(buffer, "Valid levels: ", sizeof buffer);
-	for (i = 0; i < NUMBER_OF_LEVELS; i++)
-	{
-		strlcat(buffer, levels[i], sizeof buffer);
-		strlcat(buffer, " ", sizeof buffer);
-	}
-	say("%s", buffer);
-}
-
-/*
- * bits_to_lastlog_level: converts the bitmap of lastlog levels into a nice
- * string format.  Note that this uses the global buffer, so watch out 
- */
-char	*bits_to_lastlog_level (int level)
-{
-	static	char	buffer[256]; /* this *should* be enough for this */
-	int	i,
-		p;
-
-	if (level == LOG_ALL)
-		strlcpy(buffer, "ALL", sizeof buffer);
-	else if (level == 0)
-		strlcpy(buffer, "NONE", sizeof buffer);
-	else
-	{
-		*buffer = '\0';
-		for (i = 0, p = 1; i < NUMBER_OF_LEVELS; i++, p *= 2)
-		{
-			if (level & p)
-			{
-				if (*buffer)
-					strlcat(buffer, " ", sizeof buffer);
-				strlcat(buffer, levels[i], sizeof buffer);
-			}
-		}
-	}
-	return (buffer);
-}
-
-int	parse_lastlog_level (const char *orig)
-{
-	char	*ptr,
-		*rest;
-	int	len,
-		i,
-		p,
-		level,
-		neg;
-	int	warn = 0;
-	char *	str;
-
-	if (!orig)
-		return 0;		/* Whatever */
-
-	str = LOCAL_COPY(orig);
-	level = 0;
-	while ((str = next_arg(str, &rest)) != NULL)
-	{
-	    while (str)
-	    {
-		if ((ptr = strchr(str, ',')) != NULL)
-			*ptr++ = '\0';
-		if ((len = strlen(str)) != 0)
-		{
-			if (my_strnicmp(str, "ALL", len) == 0)
-				level = LOG_ALL;
-			else if (my_strnicmp(str, "NONE", len) == 0)
-				level = 0;
-			else
-			{
-				if (*str == '-')
-				{
-					str++, len--;
-					neg = 1;
-				}
-				else
-					neg = 0;
-
-				for (i = 0, p = 1; i < NUMBER_OF_LEVELS;
-						i++, p *= 2)
-				{
-					if (!my_strnicmp(str, levels[i], len))
-					{
-						if (neg)
-							level &= (LOG_ALL ^ p);
-						else
-							level |= p;
-						break;
-					}
-				}
-				if (i == NUMBER_OF_LEVELS)
-				{
-					say("Unknown level: %s", str);
-					if (!warn)
-					{
-						warn_lastlog_levels();
-						warn = 1;
-					}
-				}
-			}
-		}
-		str = ptr;
-	    }
-	    str = rest;
-	}
-	return (level);
-}
-
-/*
- * set_lastlog_level: called whenever a "SET LASTLOG_LEVEL" is done.  It
- * parses the settings and sets the lastlog_level variable appropriately.  It
+ * set_lastlog_mask: called whenever a "SET LASTLOG_LEVEL" is done.  It
+ * parses the settings and sets the lastlog_mask variable appropriately.  It
  * also rewrites the LASTLOG_LEVEL variable to make it look nice 
  */
-void	set_lastlog_level (const void *stuff)
+void	set_lastlog_mask (void *stuff)
 {
-	const char *str = (const char *)stuff;
-	lastlog_level = parse_lastlog_level(str);
-	set_string_var(LASTLOG_LEVEL_VAR, bits_to_lastlog_level(lastlog_level));
-	current_window->lastlog_level = lastlog_level;
+	VARIABLE *v;
+	const char *str;
+	char *rejects = NULL;
+
+	v = (VARIABLE *)stuff;
+	str = v->string;
+
+	if (str_to_mask(&lastlog_mask, str, &rejects))
+		standard_level_warning("/SET LASTLOG_LEVEL", &rejects);
+	malloc_strcpy(&v->string, mask_to_str(&lastlog_mask));
+
+	current_window->lastlog_mask = lastlog_mask;
 }
 
-void	set_new_server_lastlog_level (const void *stuff)
+void	set_new_server_lastlog_mask (void *stuff)
 {
-	const char *str = (const char *)stuff;
-	new_server_lastlog_level = parse_lastlog_level(str);
-	set_string_var(NEW_SERVER_LASTLOG_LEVEL_VAR, 
-			bits_to_lastlog_level(new_server_lastlog_level));
-}
+	VARIABLE *v;
+	const char *str;
+	char *rejects = NULL;
 
+	v = (VARIABLE *)stuff;
+	str = v->string;
 
-void 	remove_from_lastlog (Window *window)
-{
-	Lastlog *new_oldest;
-	Lastlog *being_removed;
-
-	if (window->lastlog_oldest)
+	if (str)
 	{
-		being_removed = window->lastlog_oldest;
-		new_oldest = being_removed->newer;
-		window->lastlog_oldest = new_oldest;
-		if (new_oldest)
-			new_oldest->older = NULL;
-		else
-			window->lastlog_newest = NULL;
-		window->lastlog_size--;
-		new_free((char **)&being_removed->msg);
-		new_free((char **)&being_removed->target);
-		new_free((char **)&being_removed);
+	    if (!new_server_lastlog_mask)
+	    {
+		new_server_lastlog_mask = (Mask *)new_malloc(sizeof(Mask));
+		mask_unsetall(new_server_lastlog_mask);
+	    }
+	    if (str_to_mask(new_server_lastlog_mask, str, &rejects))
+		standard_level_warning("/SET NEW_SERVER_LASTLOG_LEVEL", &rejects);
+	    malloc_strcpy(&v->string, mask_to_str(new_server_lastlog_mask));
 	}
 	else
-		window->lastlog_size = 0;
+	    new_free(&new_server_lastlog_mask);
+}
+
+void	set_old_server_lastlog_mask (void *stuff)
+{
+	VARIABLE *v;
+	const char *str;
+	char *rejects = NULL;
+
+	v = (VARIABLE *)stuff;
+	str = v->string;
+
+	if (str)
+	{
+	    if (!old_server_lastlog_mask)
+	    {
+		old_server_lastlog_mask = (Mask *)new_malloc(sizeof(Mask));
+		mask_unsetall(old_server_lastlog_mask);
+	    }
+	    if (str_to_mask(old_server_lastlog_mask, str, &rejects))
+		standard_level_warning("/SET OLD_SERVER_LASTLOG_LEVEL", &rejects);
+	    malloc_strcpy(&v->string, mask_to_str(old_server_lastlog_mask));
+	}
+	else
+	    new_free(&old_server_lastlog_mask);
+}
+
+Mask	real_notify_mask (void)
+{
+	return (notify_mask);
+}
+
+Mask	real_lastlog_mask (void)
+{
+	return (lastlog_mask);
+}
+
+void	set_notify_mask (void *stuff)
+{
+	VARIABLE *v;
+	const char *str;
+	char *rejects = NULL;
+
+	v = (VARIABLE *)stuff;
+	str = v->string;
+
+	if (str_to_mask(&notify_mask, str, &rejects))
+		standard_level_warning("/SET NOTIFY_LEVEL", &rejects);
+	malloc_strcpy(&v->string, mask_to_str(&notify_mask));
+
+	current_window->notify_mask = notify_mask;
+}
+
+void	set_current_window_mask (void *stuff)
+{
+	VARIABLE *v;
+	const char *str;
+	char *rejects = NULL;
+
+	v = (VARIABLE *)stuff;
+	str = v->string;
+
+	if (str_to_mask(&current_window_mask, str, &rejects))
+		standard_level_warning("/SET CURRENT_WINDOW_LEVEL", &rejects);
+	malloc_strcpy(&v->string, mask_to_str(&current_window_mask));
+}
+
+
+/**********************************************************************/
+/*
+ * add_to_lastlog: adds the line to the lastlog.  If the LASTLOG_CONVERSATION
+ * variable is on, then only those lines that are user messages (private
+ * messages, channel messages, wall's, and any outgoing messages) are
+ * recorded, otherwise, everything is recorded 
+ */
+intmax_t	add_to_lastlog (Window *window, const char *line)
+{
+	Lastlog *new_l;
+
+	if (!window)
+		window = current_window;
+
+	new_l = (Lastlog *)new_malloc(sizeof(Lastlog));
+	new_l->dead = 0;
+	new_l->refnum = global_lastlog_refnum++;
+	new_l->older = lastlog_newest;
+	new_l->newer = NULL;
+	new_l->level = who_level;
+	new_l->msg = malloc_strdup(line);
+	new_l->winref = window->refnum;
+	if (who_from)
+		new_l->target = malloc_strdup(who_from);
+	else
+		new_l->target = NULL;
+	time(&new_l->when);
+
+	/* * * */
+	if (lastlog_newest && lastlog_newest->dead)
+		panic(1, "Lastlog_newest is dead");
+
+	if (lastlog_newest)
+		lastlog_newest->newer = new_l;
+	lastlog_newest = new_l;
+
+	if (!lastlog_oldest)
+		lastlog_oldest = lastlog_newest;
+
+	if (mask_isset(&window->lastlog_mask, who_level))
+	{
+		new_l->visible = 1;
+		window->lastlog_size++;
+		trim_lastlog(window);
+	}
+	else
+		new_l->visible = 0;
+
+	/* * * */
+	return new_l->refnum;
+}
+
+
+void 	trim_lastlog (Window *window)
+{
+	Lastlog *item = NULL;
+
+	while (window->lastlog_size > window->lastlog_max)
+	{
+		item = NULL;
+		if (oldest_lastlog_for_window(&item, window->refnum))
+		{
+			if (item->visible)
+				window->lastlog_size--;
+			remove_lastlog_item(item);
+		}
+	}
 }
 
 /*
@@ -240,36 +269,22 @@ void 	remove_from_lastlog (Window *window)
  * has gotten larger than it was before, all newer lastlog entries remain.
  * If it get smaller, some are deleted from the end. 
  */
-void	set_lastlog_size (const void *stuff)
+void	set_lastlog_size (void *stuff)
 {
-	int	size = *(const int *)stuff;
-	int	i,
-		diff;
+	VARIABLE *v;
+	int	size;
 	Window	*window = NULL;
 
+	v = (VARIABLE *)stuff;
+	size = v->integer;
+
 	while (traverse_all_windows(&window))
 	{
-		if (window->lastlog_size > size)
-		{
-			diff = window->lastlog_size - size;
-			for (i = 0; i < diff; i++)
-				remove_from_lastlog(window);
-		}
 		window->lastlog_max = size;
+		trim_lastlog(window);
 	}
 }
 
-/* set_lastlog_msg_level: sets the message level for recording in the lastlog */
-int 	set_lastlog_msg_level (int level)
-{
-	int	old;
-
-	old = msg_level;
-	msg_level = level;
-	return (old);
-}
-
-
 /*
  * The /LASTLOG command:
  * Syntax:
@@ -283,6 +298,7 @@ int 	set_lastlog_msg_level (int level)
  *	-max <number>		Only show first <number> matches
  *	-skip <number>		Skip this many leading lastlog entries
  *	-number <number>	Only search this many lastlog entries
+ *	-rewrite "<expando>"	Rewrite each line with the expando.
  *	-<LEVEL>		Add <LEVEL> to "level mask"
  *	--<LEVEL>		Remove <LEVEL> from "level mask"
  *	-ALL			Add all levels to "level mask"
@@ -331,7 +347,7 @@ int 	set_lastlog_msg_level (int level)
 BUILT_IN_COMMAND(lastlog)
 {
 	int		reverse = 0;
-	int		level_mask = 0;
+	Mask		level_mask;
 	int		skip = -1;
 	int		number = INT_MAX;
 	int		max = -1;
@@ -343,11 +359,11 @@ BUILT_IN_COMMAND(lastlog)
 	Lastlog *	l;
 	Lastlog *	lastshown;
 	regex_t 	realreg;
-	regex_t *	reg = NULL;
+	regex_t *	rex = NULL;
 	int		cnt;
 	char *		arg;
 	int		header = 1;
-	int		save_level;
+	Mask		save_mask;
 	int		before = -1;
 	int		after = 0;
 	int		counter = 0;
@@ -355,11 +371,17 @@ BUILT_IN_COMMAND(lastlog)
 	const char *	separator = "----";
 	char *		outfile = NULL;
 	FILE *		outfp = NULL;
+	int		mangler = 0;
+	int		lc;
+	char *		rewrite = NULL;
+	unsigned	winref = current_window->refnum;
 
-	message_to(0);
+	lc = message_setall(0, NULL, LEVEL_OTHER);
 	cnt = current_window->lastlog_size;
-	save_level = current_window->lastlog_level;
-	current_window->lastlog_level = 0;
+	save_mask = current_window->lastlog_mask;
+	mask_unsetall(&current_window->lastlog_mask);
+	mask_unsetall(&level_mask);
+	rewrite = get_string_var(LASTLOG_REWRITE_VAR);
 
 	while ((arg = new_next_arg(args, &args)) != NULL)
 	{
@@ -406,6 +428,12 @@ BUILT_IN_COMMAND(lastlog)
 			goto bail;
 		}
 	    }
+	    else if (!my_strnicmp(arg, "-MANGLE", len))
+	    {
+		char *	x = new_next_arg(args, &args);
+
+		mangler = parse_mangle(x, mangler, NULL);
+	    }
 	    else if (!my_strnicmp(arg, "-SKIP", len))
 	    {
 		char *x = new_next_arg(args, &args);
@@ -475,38 +503,44 @@ BUILT_IN_COMMAND(lastlog)
 	    else if (!my_strnicmp(arg, "-REVERSE", len))
 		reverse = 1;
 	    else if (!my_strnicmp(arg, "-ALL", len))
-		level_mask = LOG_ALL;
+		mask_setall(&level_mask);
 	    else if (!my_strnicmp(arg, "--ALL", len))
-		level_mask = 0;
+		mask_unsetall(&level_mask);
 	    else if (!my_strnicmp(arg, "--", 2))
 	    {
-		int	i;
-		for (i = 0; i < NUMBER_OF_LEVELS; i++)
-		{
-		    if (!my_strnicmp(levels[i], arg+2, len-2))
-		    {
-			level_mask &= ~(1 << i);
-			break;
-		    }
-		}
-		if (i == NUMBER_OF_LEVELS)
+		int	i = str_to_level(arg+2);
+		if (i != -1)
+			mask_unset(&level_mask, i);
+		else
 		{
 			say("Unknown flag: %s", arg);
 			goto bail;
 		}
 	    }
+	    else if (!my_strnicmp(arg, "-REWRITE", len))
+	    {
+		rewrite = new_next_arg(args, &args);
+	    }
+	    else if (!my_strnicmp(arg, "-WINDOW", len))
+	    {
+		Window *w;
+		char *x;
+
+		x = new_next_arg(args, &args);
+		if (!(w = get_window_by_desc(x)))
+		{
+			yell("LASTLOG -WINDOW %s is not a valid window "
+				"name or number", x);
+			goto bail;
+		}
+		winref = w->refnum;
+	    }
 	    else if (!my_strnicmp(arg, "-", 1))
 	    {
-		int	i;
-		for (i = 0; i < NUMBER_OF_LEVELS; i++)
-		{
-		    if (!my_strnicmp(levels[i], arg+1, len-1))
-		    {
-			level_mask |= (1 << i);
-			break;
-		    }
-		}
-		if (i == NUMBER_OF_LEVELS)
+		int	i = str_to_level(arg+1);
+		if (i != -1)
+			mask_set(&level_mask, i);
+		else
 		{
 			say("Unknown flag: %s", arg);
 			goto bail;
@@ -573,7 +607,7 @@ BUILT_IN_COMMAND(lastlog)
 			yell("%s", errmsg);
 			goto bail;
 		}
-		reg = &realreg;
+		rex = &realreg;
 	}
 
 	if (x_debug & DEBUG_LASTLOG)
@@ -587,7 +621,7 @@ BUILT_IN_COMMAND(lastlog)
 		yell("Skip: %d", skip);
 		yell("Number: %d", number);
 		yell("Max: %d", max);
-		yell("Mask: %d", level_mask);
+		yell("Mask: %s", mask_to_str(&level_mask));
 	}
 
 	if (outfile)
@@ -626,24 +660,29 @@ BUILT_IN_COMMAND(lastlog)
 	 */
 	if (reverse == 0)
 	{
-	    int i;
+	    int i = 0;
 
-	    end = current_window->lastlog_newest;
-	    start = end;
-	    for (i = 0; start != current_window->lastlog_oldest; )
+	    for (start = end = lastlog_newest; start != lastlog_oldest; )
 	    {
-		if (!level_mask || (level_mask & start->level))
+		if (start->visible && start->winref == winref)
+		{
+		    if (mask_isnone(&level_mask) || 
+				(mask_isset(&level_mask, start->level)))
 			i++;
-		if (i == number)
+		    if (i == number)
 			break;
+		}
 		start = start->older;
 	    }
 
 	    lastshown = NULL;
 	    for (l = start; l; (void)(l && (l = l->newer)))
 	    {
-		if (show_lastlog(&l, &skip, &number, level_mask, 
-				match, reg, &max, target))
+		char *result = NULL;
+
+		if (show_lastlog(&l, &skip, &number, &level_mask, 
+				match, rex, &max, target, mangler, winref,
+				&result))
 		{
 		    if (counter == 0 && before > 0)
 		    {
@@ -679,37 +718,62 @@ BUILT_IN_COMMAND(lastlog)
 
 		if (counter)
 		{
-			file_put_it(outfp, "%s", l->msg);
+			if (rewrite)
+			{
+				unsigned char *n, vitals[10240];
+
+				snprintf(vitals, sizeof(vitals),
+					"%ld %ld %ld %ld . . . %s %s",
+						(long)l->refnum,
+						(long)l->when,
+						(long)l->winref,
+						(long)l->level,
+						l->target?l->target:".",
+						result?result:".");
+
+				n = expand_alias(rewrite, vitals);
+				file_put_it(outfp, "%s", n);
+				new_free(&n);
+			}
+			else
+				file_put_it(outfp, "%s", result);
+
 			lastshown = l;
 			counter--;
 			if (counter == 0 && before != -1 && separator)
 				show_separator = 1;
 		}
 
+		new_free(&result);
 		if (l == end)
 			break;
 	    }
 	}
 	else
 	{
-	    int i;
+	    int i = 0;
 
-	    start = current_window->lastlog_newest;
-	    end = start;
-	    for (i = 0; end != current_window->lastlog_oldest; )
+	    for (start = end = lastlog_newest; end != lastlog_oldest; )
 	    {
-		if (!level_mask || (level_mask & start->level))
+		if (end->visible && end->winref == winref)
+		{
+		    if (mask_isnone(&level_mask) || 
+				(mask_isset(&level_mask, end->level)))
 			i++;
-		if (i == number)
+		    if (i == number)
 			break;
+		}
 		end = end->older;
 	    }
 
 	    lastshown = NULL;
 	    for (l = start; l; (void)(l && (l = l->older)))
 	    {
-		if (show_lastlog(&l, &skip, &number, level_mask, 
-				match, reg, &max, target))
+		char *result = NULL;
+
+		if (show_lastlog(&l, &skip, &number, &level_mask, 
+				match, rex, &max, target, mangler, winref,
+				&result))
 		{
 		    if (counter == 0 && before > 0)
 		    {
@@ -744,12 +808,32 @@ BUILT_IN_COMMAND(lastlog)
 		}
 		if (counter)
 		{
-			file_put_it(outfp, "%s", l->msg);
+			if (rewrite)
+			{
+				unsigned char *n, vitals[10240];
+
+				snprintf(vitals, sizeof(vitals),
+					"%ld %ld %ld %ld . . . %s %s",
+						(long)l->refnum,
+						(long)l->when,
+						(long)l->winref,
+						(long)l->level,
+						l->target?l->target:".",
+						result?result:".");
+
+				n = expand_alias(rewrite, vitals);
+				file_put_it(outfp, "%s", n);
+				new_free(&n);
+			}
+			else
+				file_put_it(outfp, "%s", result);
+
 			lastshown = l;
 			counter--;
 			if (counter == 0 && before != -1 && separator)
 				show_separator = 1;
 		}
+		new_free(&result);
 		if (l == end)
 			break;
 	    }
@@ -759,11 +843,10 @@ BUILT_IN_COMMAND(lastlog)
 bail:
 	if (outfp)
 		fclose(outfp);
-	if (reg)
-		regfree(reg);
-	current_window->lastlog_level = save_level;
-	set_lastlog_msg_level(msg_level);
-	message_to(-1);
+	if (rex)
+		regfree(rex);
+	current_window->lastlog_mask = save_mask;
+	pop_message_from(lc);
 	return;
 }
 
@@ -771,8 +854,26 @@ bail:
  * This returns 1 if the current item pointed to by 'l' is something that
  * should be displayed based on the criteron provided.
  */
-static int	show_lastlog (Lastlog **l, int *skip, int *number, int level_mask, char *match, regex_t *reg, int *max, const char *target)
+static int	show_lastlog (Lastlog **l, int *skip, int *number, Mask *level_mask, char *match, regex_t *rex, int *max, const char *target, int mangler, unsigned winref, char **result)
 {
+	const char *str = NULL;
+	*result = NULL;
+
+	if ((*l)->winref != winref)
+	{
+		if (x_debug & DEBUG_LASTLOG)
+			yell("Lastlog item belongs to another window");
+		return 0;
+	}
+
+
+	if ((*l)->visible == 0)
+	{
+		if (x_debug & DEBUG_LASTLOG)
+			yell("Lastlog item is not visible");
+		return 0;
+	}
+
 	if (*skip > 0)
 	{
 		if (x_debug & DEBUG_LASTLOG)
@@ -780,35 +881,37 @@ static int	show_lastlog (Lastlog **l, int *skip, int *number, int level_mask, ch
 		(*skip)--;	/* Have not skipped enough leading records */
 		return 0;
 	}
-#if 0
-	if (*number == 0)
-	{
-		if (x_debug & DEBUG_LASTLOG)
-			yell("Number == 0");
-		*l = NULL;	/* Have iterated over the max num. records */
-		return 0;
-	}
-	if (*number > 0)
-		(*number)--;	/* Have not yet iterated over max records */
-#endif
 
-	if (level_mask && (((*l)->level & level_mask) == 0))
+	if (!mask_isnone(level_mask) && !mask_isset(level_mask, (*l)->level))
 	{
 		if (x_debug & DEBUG_LASTLOG)
-			yell("Level_mask != level ([%d] [%d])",
-				level_mask, (*l)->level);
+			yell("Level_mask != level ([%s] [%s])",
+				mask_to_str(level_mask), level_to_str((*l)->level));
 		return 0;			/* Not of proper level */
 	}
-	if (match && !wild_match(match, (*l)->msg))
+
+	if (mangler)
+	{
+		char *	output, *rresult;
+
+		rresult = new_normalize_string((*l)->msg, 1, mangler);
+		output = LOCAL_COPY(rresult);
+		new_free(&rresult);
+		str = output;
+	}
+	else
+		str = (*l)->msg;
+
+	if (match && !wild_match(match, str))
 	{
 		if (x_debug & DEBUG_LASTLOG)
-			yell("Line [%s] not matched [%s]", (*l)->msg, match);
+			yell("Line [%s] not matched [%s]", str, match);
 		return 0;			/* Pattern match failed */
 	}
-	if (reg && regexec(reg, (*l)->msg, 0, NULL, 0))
+	if (rex && regexec(rex, str, 0, NULL, 0))
 	{
 		if (x_debug & DEBUG_LASTLOG)
-			yell("Line [%s] not regexed", (*l)->msg);
+			yell("Line [%s] not regexed", str);
 		return 0;			/* Regex match failed */
 	}
 	if (target && (!(*l)->target || !wild_match(target, (*l)->target)))
@@ -828,88 +931,26 @@ static int	show_lastlog (Lastlog **l, int *skip, int *number, int level_mask, ch
 	if (*max > 0)
 		(*max)--;	/* Have not yet shown max number of matches */
 
+	(*result) = malloc_strdup(str);
 	return 1;		/* Show it! */
 }
 
 /*
- * add_to_lastlog: adds the line to the lastlog.  If the LASTLOG_CONVERSATION
- * variable is on, then only those lines that are user messages (private
- * messages, channel messages, wall's, and any outgoing messages) are
- * recorded, otherwise, everything is recorded 
+ * reconstitute_scrollback: walk through the lastlog, and put_it everything,
+ * making sure to reset the level and all that jazz.  This will cause the 
+ * scrollback to be rebroken, etc.
  */
-void 	add_to_lastlog (Window *window, const char *line)
+void	reconstitute_scrollback (Window *window)
 {
-	Lastlog *new_l;
+	Lastlog *li;
 
-	if (!window)
-		window = current_window;
-
-	if (window->lastlog_level & msg_level)
+	for (li = lastlog_oldest; li; li = li->newer)
 	{
-		new_l = (Lastlog *)new_malloc(sizeof(Lastlog));
-		new_l->older = window->lastlog_newest;
-		new_l->newer = NULL;
-		new_l->level = msg_level;
-		new_l->msg = malloc_strdup(line);
-		if (who_from)
-			new_l->target = malloc_strdup(who_from);
-		else
-			new_l->target = NULL;
-
-		if (window->lastlog_newest)
-			window->lastlog_newest->newer = new_l;
-		window->lastlog_newest = new_l;
-
-		if (!window->lastlog_oldest)
-			window->lastlog_oldest = window->lastlog_newest;
-
-		if (window->lastlog_size++ >= window->lastlog_max)
-			remove_from_lastlog(window);
+	    if (li->winref == window->refnum)
+		add_to_window_scrollback(window, li->msg, li->refnum);
 	}
 }
-
-int	real_notify_level (void)
-{
-	return (notify_level);
-}
-
-int	real_lastlog_level (void)
-{
-	return (lastlog_level);
-}
-
-void	set_notify_level (const void *stuff)
-{
-	const char *str = (const char *)stuff;
-	notify_level = parse_lastlog_level(str);
-	set_string_var(NOTIFY_LEVEL_VAR, bits_to_lastlog_level(notify_level));
-	current_window->notify_level = notify_level;
-}
-
-void 	set_beep_on_msg (const void *stuff)
-{
-	const char *str = (const char *)stuff;
-	beep_on_level = parse_lastlog_level(str);
-	set_string_var(BEEP_ON_MSG_VAR, bits_to_lastlog_level(beep_on_level));
-}
-
-void	set_current_window_level (const void *stuff)
-{
-	const char *str = (const char *)stuff;
-	current_window_level = parse_lastlog_level(str);
-	set_string_var(CURRENT_WINDOW_LEVEL_VAR, 
-			bits_to_lastlog_level(current_window_level));
-}
-
-#if 0
-#define EMPTY empty_string
-#define RETURN_EMPTY return malloc_strdup(EMPTY)
-#define RETURN_IF_EMPTY(x) if (empty( x )) RETURN_EMPTY
-#define GET_INT_ARG(x, y) {RETURN_IF_EMPTY(y); x = my_atol(safe_new_next_arg(y, &y));}
-#define GET_STR_ARG(x, y) {RETURN_IF_EMPTY((y)); x = new_next_arg((y), &(y));RETURN_IF_EMPTY((x));}
-#define RETURN_STR(x) return malloc_strdup(x ? x : EMPTY);
-#endif
-
+	
 /*
  * $line(<line number> [window number])
  * Returns the text of logical line <line number> from the lastlog of 
@@ -920,23 +961,28 @@ void	set_current_window_level (const void *stuff)
  * Lines are numbered from 1, starting at the most recent line in the buffer.
  * Contributed by Crackbaby (Matt Carothers) on March 19, 1998.
  */
-char 	*function_line (char *word)
+BUILT_IN_FUNCTION(function_line, word)
 {
 	int	line = 0;
-	char *	windesc = zero;
+	const char *	windesc = zero;
 	Lastlog	*start_pos;
 	Window	*win;
 	char	*extra;
 	int	do_level = 0;
+	int	do_timestamp = 0;
+	char *	retval = NULL;
+	size_t	clue = 0;
 
 	GET_INT_ARG(line, word);
 
 	while (word && *word)
 	{
-		GET_STR_ARG(extra, word);
+		GET_FUNC_ARG(extra, word);
 
 		if (!my_stricmp(extra, "-LEVEL"))
 			do_level = 1;
+		else if (!my_stricmp(extra, "-TIME"))
+			do_timestamp = 1;
 		else
 			windesc = extra;
 	}
@@ -950,19 +996,37 @@ char 	*function_line (char *word)
 		RETURN_EMPTY;
 
 	/* Get the line from the lastlog */
-	for (start_pos = win->lastlog_newest; line; start_pos = start_pos->older)
-		line--;
+	for (start_pos = lastlog_newest; line; start_pos = start_pos->older)
+	{
+		if (start_pos->winref != win->refnum)
+			continue;
+		if (start_pos->visible)
+			line--;
+	}
 
 	if (!start_pos)
-		start_pos = win->lastlog_oldest;
+		start_pos = lastlog_oldest;
 	else
 		start_pos = start_pos->newer;
 
+	while ((start_pos->visible == 0 || start_pos->winref != win->refnum) && 
+				start_pos->newer)
+		start_pos = start_pos->newer;
+
+	/* If there are no visible lastlog items, punt */
+	if (!start_pos)
+		RETURN_EMPTY;
+
+	malloc_strcat_c(&retval, start_pos->msg, &clue);
+
 	if (do_level)
-		return malloc_sprintf(NULL, "%s %s", start_pos->msg, 
-					levels[start_pos->level]);
-	else
-		RETURN_STR(start_pos->msg);
+		malloc_strcat_wordlist_c(&retval, space, 
+					level_to_str(start_pos->level), &clue);
+	if (do_timestamp)
+		malloc_strcat_wordlist_c(&retval, space, 
+					NUMSTR(start_pos->when), &clue);
+
+	RETURN_MSTR(retval);
 }
 
 /*
@@ -972,30 +1036,39 @@ char 	*function_line (char *word)
  * levels as represented by the lastlog levels. If the window number is 
  * invalid, the function will return the false value.
  */
-char *function_lastlog (char *word)
+BUILT_IN_FUNCTION(function_lastlog, word)
 {
-	char *	windesc = zero;
+	const char *	windesc = zero;
 	char *	pattern = NULL;
 	char *	retval = NULL;
 	Lastlog	*iter;
 	Window *win;
-	int	lastlog_levels;
+	Mask	lastlog_levels;
 	int	line = 1;
 	size_t	rvclue = 0;
+	char *	rejects = NULL;
 
-	GET_STR_ARG(windesc, word);
-	GET_STR_ARG(pattern, word);
-	lastlog_levels = parse_lastlog_level(word);
+	GET_FUNC_ARG(windesc, word);
+	GET_DWORD_ARG(pattern, word);
+	if (str_to_mask(&lastlog_levels, word, &rejects))
+		standard_level_warning("$lastlog()", &rejects);
 
 	/* Get the current window, default to current window */
 	if (!(win = get_window_by_desc(windesc)))
 		RETURN_EMPTY;
 
-	for (iter = win->lastlog_newest; iter; iter = iter->older, line++)
+	for (iter = lastlog_newest; iter; iter = iter->older)
 	{
-		if (iter->level & lastlog_levels)
-			if (wild_match(pattern, iter->msg))
-				malloc_strcat_word_c(&retval, space, ltoa(line), &rvclue);
+		if (iter->winref != win->refnum)
+			continue;
+		if (iter->visible == 0)
+			continue;
+
+		if (mask_isset(&lastlog_levels, iter->level))
+		    if (wild_match(pattern, iter->msg))
+			malloc_strcat_word_c(&retval, space, 
+					ltoa(line), DWORD_NO, &rvclue);
+		line++;
 	}
 
 	if (retval)
@@ -1004,3 +1077,244 @@ char *function_lastlog (char *word)
 	RETURN_EMPTY;
 }
 
+
+#if 0
+/*
+ * Here's the plan:
+ * 	$lastlogctl(REFNUMS <windesc> <levels>)
+ *	  Like $lastlog(), returns all lastlog refnums for <windesc> that
+ *	  belong to one of the level(s).  The refnums returned do not work
+ *	  with $line() (but work with $lastlogctl())
+ *
+ *	$lastlogctl(GET <refnum> LEVEL)
+ *	  The lastlog item's level, eg, MSGS, PUBLICS, OTHER.
+ *	$lastlogctl(GET <refnum> TARGET)
+ *	  The item's target: 
+ *		For messages sent to you, the sender.
+ *	  	For messages you sent, the receiver.  
+ *		For messages sent to a channel, the channel.
+ *		For everything else, the empty string.
+ *	$lastlogctl(GET <refnum> TEXT)
+ *	$lastlogctl(GET <refnum> TIMESTAMP)
+ *	$lastlogctl(GET <refnum> VISIBLE)
+ *	$lastlogctl(GET <refnum> WINDOW)
+ *
+ *	$lastlogctl(SET <refnum> LEVEL <level>)
+ *	$lastlogctl(SET <refnum> TARGET <target>)
+ *	$lastlogctl(SET <refnum> TEXT <text>)
+ *	$lastlogctl(SET <refnum> TIMESTAMP <timestamp>)
+ *	$lastlogctl(SET <refnum> VISIBLE [0|1])
+ *	$lastlogctl(SET <refnum> WINDOW)
+ */
+BUILT_IN_FUNCTION(function_lastlogctl, input)
+{
+	RETURN_STR(empty_string);
+}
+#endif
+
+/************************************************************************/
+
+static int	oldest_lastlog_for_window (Lastlog **item, unsigned winref)
+{
+	*item = NULL;
+	return newer_lastlog_entry(item, winref);
+}
+
+static int	newer_lastlog_entry (Lastlog **item, unsigned winref)
+{
+	Lastlog *i;
+
+	if (*item)
+		i = (*item)->newer;
+	else
+		i = lastlog_oldest;
+
+	while (i && i->winref != winref)
+		i = i->newer;
+	*item = i;
+
+	return i ? 1 : 0;
+}
+
+static int	older_lastlog_entry (Lastlog **item, unsigned winref)
+{
+	Lastlog *i;
+
+	if (*item)
+		i = (*item)->older;
+	else
+		i = lastlog_newest;
+
+	while (i && i->winref != winref)
+		i = i->older;
+	*item = i;
+
+	return i ? 1 : 0;
+}
+
+static int	newest_lastlog_for_window (Lastlog **item, unsigned winref)
+{
+	*item = NULL;
+	return older_lastlog_entry(item, winref);
+}
+
+/*
+ * Because this function does not have access to the window that this
+ * lastlog item belongs to *the caller* is responsible for adjusting
+ * window->lastlog_size!  I could change that in the future I guess
+ */
+static void	remove_lastlog_item (Lastlog *item)
+{
+	if (item->dead)
+		panic(1, "Lastlog item is already dead.");
+
+	if (item == lastlog_oldest)
+	{
+		if (item->older != NULL)
+			panic(1, "Oldest lastlog item %jd has older item %jd",
+				item->refnum, item->older->refnum);
+
+		if (item->newer)
+			item->newer->older = NULL;
+		lastlog_oldest = item->newer;
+		item->newer = NULL;
+	}
+
+	if (item == lastlog_newest)
+	{
+		if (item->newer != NULL)
+			panic(1, "Newest lastlog item %jd has newer item %jd",
+				item->refnum, item->newer->refnum);
+
+		if (item->older)
+			item->older->newer = NULL;
+		lastlog_newest = item->older;
+		item->older = NULL;
+	}
+
+	if (item->older)
+		item->older->newer = item->newer;
+	if (item->newer)
+		item->newer->older = item->older;
+	item->newer = item->older = NULL;
+
+	item->dead = 1;
+	new_free((char **)&item->msg);
+	new_free((char **)&item->target);
+	new_free((char **)&item);
+}
+
+static void	switch_lastlog_window (Lastlog *item, unsigned newref)
+{
+	/* Mark the old window's scrollback for reconstitution */
+	window_scrollback_needs_rebuild(item->winref);
+	item->winref = newref;
+	/* Mark the new window's scrollback for reconstitution */
+	window_scrollback_needs_rebuild(item->winref);
+}
+
+/***************************************************************************/
+static void	move_lastlog_item (Lastlog *item, unsigned newref)
+{
+	unsigned	oldref = item->winref;
+
+	item->winref = newref;
+	window_scrollback_needs_rebuild(oldref);
+	window_scrollback_needs_rebuild(newref);
+}
+
+void	move_all_lastlog (unsigned oldref, unsigned newref)
+{
+	Lastlog *l;
+
+	for (l = lastlog_oldest; l; l = l->newer)
+	{
+		if (l->winref != oldref)
+			continue;
+		move_lastlog_item(l, newref);
+	}
+}
+
+void	move_lastlog_item_by_string (unsigned oldref, unsigned newref, const char *str)
+{
+	Lastlog *l;
+
+	for (l = lastlog_oldest; l; l = l->newer)
+	{
+		if (l->winref != oldref)
+			continue;
+		if (!stristr(l->msg, str))
+			continue;
+		move_lastlog_item(l, newref);
+	}
+}
+
+void	move_lastlog_item_by_target (unsigned oldref, unsigned newref, const char *str)
+{
+	Lastlog *l;
+
+	for (l = lastlog_oldest; l; l = l->newer)
+	{
+		if (l->winref != oldref)
+			continue;
+		if (!my_stricmp(l->target, str))
+			continue;
+		move_lastlog_item(l, newref);
+	}
+}
+
+void	move_lastlog_item_by_level (unsigned oldref, unsigned newref, Mask *levels)
+{
+	Lastlog *l;
+
+	for (l = lastlog_oldest; l; l = l->newer)
+	{
+		if (l->winref != oldref)
+			continue;
+		if (!mask_isset(levels, l->level))
+			continue;
+		move_lastlog_item(l, newref);
+	}
+}
+
+void	move_lastlog_item_by_regex (unsigned oldref, unsigned newref, const char *str)
+{
+	Lastlog *l;
+	regex_t preg;
+	int	errcode;
+
+	errcode = regcomp(&preg, str, REG_EXTENDED | REG_ICASE | REG_NOSUB);
+	if (errcode != 0)
+	{
+		char	errstr[256];
+		regerror(errcode, &preg, errstr, sizeof(errstr));
+		say("Regular expression [%s] does not compile: %s", 
+			str, errstr);
+		return;
+	}
+
+	for (l = lastlog_oldest; l; l = l->newer)
+	{
+		if (l->winref != oldref)
+			continue;
+		if (regexec(&preg, l->msg, 0, NULL, 0))
+			continue;
+		move_lastlog_item(l, newref);
+	}
+
+	regfree(&preg);
+}
+
+void	lastlog_swap_winrefs (unsigned oldref, unsigned newref)
+{
+	Lastlog *l;
+
+	for (l = lastlog_oldest; l; l = l->newer)
+	{
+		if (l->winref == oldref)
+			l->winref = newref;
+		else if (l->winref == newref)
+			l->winref = oldref;
+	}
+}
+
