@@ -1,4 +1,4 @@
-/* $EPIC: logfiles.c,v 1.24 2005/04/10 14:58:41 jnelson Exp $ */
+/* $EPIC: logfiles.c,v 1.37 2010/02/15 03:59:10 jnelson Exp $ */
 /*
  * logfiles.c - General purpose log files
  *
@@ -31,6 +31,7 @@
  * SUCH DAMAGE.
  */
 #include "irc.h"
+#include "levels.h"
 #include "log.h"
 #include "vars.h"
 #include "output.h"
@@ -61,7 +62,7 @@ struct Logfile {
 	int	type;
 	WNickList *targets;
 	int	refnums[MAX_TARGETS];
-	int	level;
+	Mask	mask;
 
 	char *	rewrite;
 	int	mangler;
@@ -77,11 +78,20 @@ int	logref = 0;
 
 static Logfile *	new_logfile (void)
 {
-	Logfile *log;
+	Logfile *log, *ptr;
 	int	i;
 
 	log = (Logfile *)new_malloc(sizeof(Logfile));
-	log->next = logfiles;
+
+	/* Move it to the end of the list. */
+	for (ptr = logfiles; ptr && ptr->next;)
+		ptr = ptr->next;
+	if (ptr)
+		ptr->next = log;
+	else
+		logfiles = log;
+
+	log->next = NULL;
 	log->refnum = ++logref;
 	log->name = malloc_sprintf(NULL, "%d", log->refnum);
 	log->filename = NULL;
@@ -91,14 +101,13 @@ static Logfile *	new_logfile (void)
 	log->targets = NULL;
 	for (i = 0; i < MAX_TARGETS; i++)
 		log->refnums[i] = -1;
-	log->level = LOG_ALL;
+	mask_setall(&log->mask);
 	log->rewrite = NULL;
 	log->mangler = 0;
 	log->mangle_desc = NULL;
 	log->active = 0;
 	time(&log->activity);
 
-	logfiles = log;
 	return log;
 }
 
@@ -270,7 +279,13 @@ static Logfile *	logfile_add (Logfile *log, char **args)
 		}
 		else if (log->type == LOG_SERVERS || log->type == LOG_WINDOWS)
 		{
-		    int refnum = my_atol(arg);
+		    int refnum;
+
+		    if (log->type == LOG_SERVERS && !my_strnicmp("ALL", arg, 1))
+			refnum = NOSERV;
+		    else
+			refnum = my_atol(arg);
+
 		    for (i = 0; i < MAX_TARGETS; i++)
 		    {
 			if (log->refnums[i] == refnum)
@@ -313,9 +328,12 @@ static Logfile *	logfile_describe (Logfile *log, char **args)
 	say("\t Logical name: %s", log->name);
 	say("\t     Filename: %s", log->filename ? log->filename : "<NONE>");
 	say("\t         Type: %s", logtype[log->type]);
-	say("\t       Server: %d", log->servref);
+	if (log->servref == NOSERV)
+		say("\t       Server: ALL");
+	else
+		say("\t       Server: %d", log->servref);
 	say("\tTarget/Refnum: %s", targets ? targets : "<NONE>");
-	say("\t        Level: %s", bits_to_lastlog_level(log->level));
+	say("\t        Level: %s", mask_to_str(&log->mask));
 	say("\t Rewrite Rule: %s", log->rewrite ? log->rewrite : "<NONE>");
 	say("\t Mangle rules: %s", log->mangle_desc ? log->mangle_desc : "<NONE>");
 
@@ -367,6 +385,7 @@ static Logfile *	logfile_kill (Logfile *log, char **args)
 static Logfile *	logfile_level (Logfile *log, char **args)
 {
         char *arg = new_next_arg(*args, args);
+	char *rejects = NULL;
 
 	if (!log)
 	{
@@ -374,7 +393,8 @@ static Logfile *	logfile_level (Logfile *log, char **args)
 		return NULL;
 	}
 
-	log->level = parse_lastlog_level(arg);
+	if (str_to_mask(&log->mask, arg, &rejects))
+		standard_level_warning("/LOG LEVEL", &rejects);
 	return log;
 }
 
@@ -387,11 +407,11 @@ static Logfile *	logfile_list (Logfile *log, char **args)
 	for (l = logfiles; l; l = l->next)
 	{
 		targets = logfile_get_targets(l);
-		say("Log %2d [%s] logging %s is %s, file %s server %d targets %s",
+		say("Log %2d [%s] logging %s is %s, file %s server %s targets %s",
 			l->refnum, l->name, logtype[l->type],
 			onoff[l->active],
 			l->filename ? l->filename : "<NONE>", 
-			l->servref,
+			l->servref == NOSERV ?  "ALL" : ltoa(l->servref),
 			targets ? targets : "<NONE>");
 		new_free(&targets);
 	}
@@ -583,10 +603,12 @@ static Logfile *	logfile_server (Logfile *log, char **args)
 		return NULL;
 	}
 
-	if (!is_number(arg))
-		say("SERVER: The log's server needs to be a number");
+	if (!my_strnicmp(arg, "ALL", 1))
+		log->servref = NOSERV;
+	else if (!is_number(arg))
+		say("/LOG SERVER: The log's server needs to be a number or ALL");
 	else
-		log->servref = parse_server_index(arg, 1);
+		log->servref = str_to_servref(arg);
 
 	return log;
 }
@@ -693,10 +715,10 @@ void	add_to_logs (long winref, int servref, const char *target, int level, const
 	    {
 		for (i = 0; i < MAX_TARGETS; i++) {
 		    if (log->refnums[i] == winref) {
-			if (log->level && (log->level & level) == 0)
+			if (!mask_isset(&log->mask, level))
 				continue;
 			time(&log->activity);
-			add_to_log(log->log, winref, orig_str, log->mangler, log->rewrite);
+			add_to_log(log->refnum, log->log, winref, orig_str, log->mangler, log->rewrite);
 		    }
 		}
 	    }
@@ -704,11 +726,13 @@ void	add_to_logs (long winref, int servref, const char *target, int level, const
 	    if (log->type == LOG_SERVERS)
 	    {
 		for (i = 0; i < MAX_TARGETS; i++) {
-		    if (log->refnums[i] == servref) {
-			if (log->level && (log->level & level) == 0)
+		    if (log->refnums[i] == NOSERV ||
+			log->refnums[i] == servref) 
+		    {
+			if (!mask_isset(&log->mask, level))
 				continue;
 			time(&log->activity);
-			add_to_log(log->log, winref, orig_str, log->mangler, log->rewrite);
+			add_to_log(log->refnum, log->log, winref, orig_str, log->mangler, log->rewrite);
 		    }
 		}
 	    }
@@ -718,7 +742,7 @@ void	add_to_logs (long winref, int servref, const char *target, int level, const
 		if (log->servref != NOSERV && (log->servref != servref))
 			continue;
 
-		if (log->level && (log->level & level) == 0)
+		if (!mask_isset(&log->mask, level))
 			continue;
 
 		if (log->targets && !target)
@@ -729,7 +753,7 @@ void	add_to_logs (long winref, int servref, const char *target, int level, const
 
 		/* OK!  We want to log it now! */
 		time(&log->activity);
-		add_to_log(log->log, winref, orig_str, log->mangler, log->rewrite);
+		add_to_log(log->refnum, log->log, winref, orig_str, log->mangler, log->rewrite);
 	    }
 	}
 }
@@ -765,12 +789,12 @@ char *logctl	(char *input)
 	int	val;
 	Logfile	*log;
 
-	GET_STR_ARG(listc, input);
+	GET_FUNC_ARG(listc, input);
         if (!my_strnicmp(listc, "REFNUMS", 7)) {
 		char *	retval = NULL;
 		int	active;
 
-		GET_STR_ARG(refstr, input);
+		GET_FUNC_ARG(refstr, input);
 		if (!my_stricmp(refstr, "ACTIVE"))
 			active = 1;
 		else if (!my_stricmp(refstr, "INACTIVE"))
@@ -784,32 +808,32 @@ char *logctl	(char *input)
 		{
 			if (active != -1 && active != log->active)
 				continue;
-			malloc_strcat_word(&retval, space, ltoa(log->refnum));
+			malloc_strcat_word(&retval, space, ltoa(log->refnum), DWORD_NO);
 		}
 		RETURN_MSTR(retval);
         } else if (!my_strnicmp(listc, "REFNUM", 6)) {
-		GET_STR_ARG(refstr, input);
+		GET_FUNC_ARG(refstr, input);
 		if (!(log = get_log_by_desc(refstr)))
 			RETURN_EMPTY;
 		RETURN_INT(log->refnum);
         } else if (!my_strnicmp(listc, "ADD", 2)) {
-		GET_STR_ARG(refstr, input);
+		GET_FUNC_ARG(refstr, input);
 		if (!(log = get_log_by_desc(refstr)))
 			RETURN_EMPTY;
 		logfile_add(log, &input);
 		RETURN_INT(1);
         } else if (!my_strnicmp(listc, "DELETE", 2)) {
-		GET_STR_ARG(refstr, input);
+		GET_FUNC_ARG(refstr, input);
 		if (!(log = get_log_by_desc(refstr)))
 			RETURN_EMPTY;
 		logfile_remove(log, &input);
 		RETURN_INT(1);
         } else if (!my_strnicmp(listc, "GET", 2)) {
-                GET_STR_ARG(refstr, input);
+                GET_FUNC_ARG(refstr, input);
 		if (!(log = get_log_by_desc(refstr)))
 			RETURN_EMPTY;
 
-                GET_STR_ARG(listc, input);
+                GET_FUNC_ARG(listc, input);
                 if (!my_strnicmp(listc, "REFNUM", 1)) {
 			RETURN_INT(log->refnum);
                 } else if (!my_strnicmp(listc, "NAME", 3)) {
@@ -822,7 +846,7 @@ char *logctl	(char *input)
 			char *ret = logfile_get_targets(log);
 			RETURN_MSTR(ret);
                 } else if (!my_strnicmp(listc, "LEVEL", 3)) {
-			char *ret = bits_to_lastlog_level(log->level);
+			const char *ret = mask_to_str(&log->mask);
 			RETURN_STR(ret);
                 } else if (!my_strnicmp(listc, "REWRITE", 3)) {
 			RETURN_STR(log->rewrite);
@@ -836,11 +860,11 @@ char *logctl	(char *input)
 			RETURN_INT(log->activity);
 		}
         } else if (!my_strnicmp(listc, "SET", 1)) {
-                GET_STR_ARG(refstr, input);
+                GET_FUNC_ARG(refstr, input);
 		if (!(log = get_log_by_desc(refstr)))
 			RETURN_EMPTY;
 
-		GET_STR_ARG(listc, input);
+		GET_FUNC_ARG(listc, input);
                 if (!my_strnicmp(listc, "NAME", 3)) {
 			logfile_name(log, &input);
 			RETURN_INT(1);
@@ -881,9 +905,35 @@ char *logctl	(char *input)
                 RETURN_EMPTY;           /* Not implemented for now. */
         } else if (!my_strnicmp(listc, "PMATCH", 1)) {
                 RETURN_EMPTY;           /* Not implemented for now. */
+        } else if (!my_strnicmp(listc, "CURRENT", 1)) {
+		RETURN_INT(current_log_refnum);
         } else
                 RETURN_EMPTY;
 
         RETURN_EMPTY;
 }
 
+/*
+ * The /WINDOW NUMBER command actually swaps the refnums of two windows:
+ * It's possible that 'newref' isn't in use, so that's ok.
+ */
+void    logfiles_swap_winrefs (int oldref, int newref)
+{
+	Logfile *log;
+	int	i;
+
+	for (log = logfiles; log; log = log->next)
+        {
+		if (log->type != LOG_WINDOWS)
+			continue;
+
+		for (i = 0; i < MAX_TARGETS; i++)
+		{
+			if (log->refnums[i] == newref)
+				log->refnums[i] = oldref;
+			else if (log->refnums[i] == oldref)
+				log->refnums[i] = newref;
+		}
+        }
+}
+
