@@ -1,4 +1,4 @@
-/* $EPIC: timer.c,v 1.42 2004/12/01 23:47:51 jnelson Exp $ */
+/* $EPIC: timer.c,v 1.55 2010/04/02 23:14:45 jnelson Exp $ */
 /*
  * timer.c -- handles timers in ircII
  *
@@ -54,7 +54,7 @@
 	int 	timer_exists (const char *ref);
 	int 	remove_timer (const char *ref);
 static void 	remove_all_timers (void);
-static	void	remove_window_timers (int winref);
+static	void	remove_timers_by_domref (int domain, int domref);
 static	void	list_timers (const char *command);
 
 /*
@@ -67,124 +67,167 @@ BUILT_IN_COMMAND(timercmd)
 {
 	char	*waittime,
 		*flag;
-	char	*want = empty_string;
+	const char	*want = empty_string;
 	char	*ptr;
 	double	interval;
 	long	events = -2;
 	int	update = 0;
-	int	winref;
 	size_t	len;
+	TimerDomain	domain;
+	int	domref;
+	int	cancelable = 0;
 
-	winref = current_window ? (int)current_window->refnum : -1;
+	if (parsing_server_index != NOSERV)
+	{
+		domain = SERVER_TIMER;
+		domref = parsing_server_index;
+	}
+	else
+	{
+		domain = WINDOW_TIMER;
+		domref = current_window ? (int)current_window->refnum : -1;
+	}
 
 	while (*args == '-' || *args == '/')
 	{
-		flag = next_arg(args, &args);
-		len = strlen(flag + 1);
+	    flag = next_arg(args, &args);
+	    len = strlen(flag + 1);
 
-		if (!my_strnicmp(flag + 1, "DELETE", len))
+	    if (!my_strnicmp(flag + 1, "DELETE", len))
+	    {
+		if (!(ptr = next_arg(args, &args)))
+		    say("%s: Need a timer reference number for -DELETE",
+				command);
+
+		/* 
+		 * Try to delete what was given.  If that
+		 * doesn't work, see if it is "delete all".
+		 * Delete_timer returns -1 on error.
+		 */
+		else if (timer_exists(ptr))
+			remove_timer(ptr);
+
+		/*
+		 * Check to see if it is /timer -delete all
+		 *
+		 * Ugh.  Yes, this hack really _is_ required.
+		 * Please don't change it without talking to
+		 * me first to understand why.
+		 */
+		else if (*ptr && !my_strnicmp(ptr, "ALL", strlen(ptr)))
+			remove_all_timers();
+		else if (*ptr)
+			say("The timer \"%s\" doesn't exist!", ptr);
+		else
+			panic(1, "how did timercmd get here?");
+
+		return;
+	    }
+	    else if (!my_strnicmp(flag + 1, "DELETE_FOR_WINDOW", len))
+	    {
+		if (!(ptr = next_arg(args, &args)) || !is_number(ptr))
 		{
-			if (!(ptr = next_arg(args, &args)))
-			    say("%s: Need a timer reference number for -DELETE",
-					command);
-
-			/* 
-			 * Try to delete what was given.  If that
-			 * doesn't work, see if it is "delete all".
-			 * Delete_timer returns -1 on error.
-			 */
-			else if (timer_exists(ptr))
-				remove_timer(ptr);
-
-			/*
-			 * Check to see if it is /timer -delete all
-			 *
-			 * Ugh.  Yes, this hack really _is_ required.
-			 * Please don't change it without talking to
-			 * me first to understand why.
-			 */
-			else if (*ptr && !my_strnicmp(ptr, "ALL", strlen(ptr)))
-				remove_all_timers();
-			else if (*ptr)
-				say("The timer \"%s\" doesn't exist!", ptr);
-			else
-				panic("how did timercmd get here?");
+		    say("%s: Need a window number for -DELETE_FOR_WINDOW", 
+				command);
+		    return;
+		}
 
+		domref = atol(ptr);
+		remove_timers_by_domref(WINDOW_TIMER, domref);
+		return;
+	    }
+	    else if (!my_strnicmp(flag+1, "REFNUM", 3))	/* REFNUM */
+	    {
+		want = next_arg(args, &args);
+		if (!want || !*want)
+		{
+			say("%s: Missing argument to -REFNUM", command);
 			return;
 		}
-		else if (!my_strnicmp(flag + 1, "DELETE_FOR_WINDOW", len))
+
+		continue;
+	    }
+	    else if (!my_strnicmp(flag+1, "REPEAT", 3))	/* REPEAT */
+	    {
+		char *na = next_arg(args, &args);
+		if (!na || !*na)
 		{
-			if (!(ptr = next_arg(args, &args)) || !is_number(ptr))
+			say("%s: Missing argument to -REPEAT", command);
+			return;
+		}
+		if (!strcmp(na, "*") || !strcmp(na, "-1"))
+			events = -1;
+		else if ((events = my_atol(na)) == 0)
+			return;	 /* Repeating 0 times is a noop */
+
+		continue;
+	    }
+	    else if (!my_strnicmp(flag + 1, "UPDATE", 1))	/* UPDATE */
+		update = 1;
+
+	    else if (!my_strnicmp(flag + 1, "LIST", 1))	/* LIST */
+	    {
+		list_timers(command);
+		return;
+	    }
+	    else if (!my_strnicmp(flag + 1, "WINDOW", 1))	/* WINDOW */
+	    {
+		char 	*na;
+		Window *win = NULL;
+
+		domain = WINDOW_TIMER;
+		if ((na = next_arg(args, &args)))
+			win = get_window_by_desc(na);
+
+		if (!win)
+		{
+		    if (my_stricmp(na, "-1"))
+		    {
+			say("%s: That window doesnt exist!", command);
+			return;
+		    }
+		    else
+			domref = -1;
+		}
+		else
+			domref = win->refnum;
+	    }
+	    else if (!my_strnicmp(flag + 1, "SERVER", 1))	/* SERVER */
+	    {
+		char 	*na;
+
+		domain = SERVER_TIMER;
+		domref = from_server;
+
+		if ((na = next_arg(args, &args)))
+		{
+		    if ((domref = str_to_servref(na)) == NOSERV)
+		    {
+			/* -1 defaults to from_server anyways */
+			if (!my_stricmp(na, "-1"))
+			    domref = from_server;
+			else
 			{
-			    say("%s: Need a window number for -DELETE_FOR_WINDOW", command);
+			    say("%s: Server %s doesnt exist!", command, na);
 			    return;
 			}
-
-			winref = atol(ptr);
-			remove_window_timers(winref);
-			return;
-		}
-		else if (!my_strnicmp(flag+1, "REF", 3))	/* REFNUM */
-		{
-			want = next_arg(args, &args);
-			if (!want || !*want)
-			{
-				say("%s: Missing argument to -REFNUM", command);
-				return;
-			}
-
-			continue;
-		}
-		else if (!my_strnicmp(flag+1, "REP", 3))	/* REPEAT */
-		{
-			char *na = next_arg(args, &args);
-			if (!na || !*na)
-			{
-				say("%s: Missing argument to -REPEAT", command);
-				return;
-			}
-			if (!strcmp(na, "*") || !strcmp(na, "-1"))
-				events = -1;
-			else if ((events = my_atol(na)) == 0)
-				return;	 /* Repeating 0 times is a noop */
-
-			continue;
-		}
-		else if (!my_strnicmp(flag + 1, "U", 1))	/* UPDATE */
-			update = 1;
-
-		else if (!my_strnicmp(flag + 1, "L", 1))	/* LIST */
-		{
-			list_timers(command);
-			return;
-		}
-		else if (!my_strnicmp(flag + 1, "W", 1))	/* WINDOW */
-		{
-			char 	*na;
-			Window *win = NULL;
-
-			if ((na = next_arg(args, &args)))
-				win = get_window_by_desc(na);
-
-			if (!win)
-			{
-			    if (my_stricmp(na, "-1"))
-			    {
-				say("%s: That window doesnt exist!", command);
-				return;
-			    }
-			    else
-				winref = -1;
-			}
-			else
-				winref = win->refnum;
-		}
-
-		else
-		{
-			say("%s: %s no such flag", command, flag);
-			return;
+		    }
 		}
+	    }
+	    else if (!my_strnicmp(flag + 1, "GENERAL", 1))	/* GENERAL */
+	    {
+		domain = GENERAL_TIMER;
+		domref = -1;
+	    }
+	    else if (!my_strnicmp(flag + 1, "CANCELABLE", 1))	/* CANCELABLE */
+	    {
+		cancelable = 1;
+	    }
+	    else
+	    {
+		say("%s: %s no such flag", command, flag);
+		return;
+	    }
 	}
 
 
@@ -198,7 +241,8 @@ BUILT_IN_COMMAND(timercmd)
 				update = 0;
 			else
 			{
-				say("%s: To use -UPDATE you must specify a valid refnum", command);
+				say("%s: To use -UPDATE you must specify a "
+						"valid refnum", command);
 				return;
 			}
 		}
@@ -215,7 +259,8 @@ BUILT_IN_COMMAND(timercmd)
 			events = -1;
 */
 
-		add_timer(update, want, interval, events, NULL, args, subargs, winref);
+		add_timer(update, want, interval, events, NULL, args, subargs, 
+				domain, domref, cancelable);
 	}
 	else
 		list_timers(command);
@@ -230,7 +275,7 @@ BUILT_IN_COMMAND(timercmd)
 #define REFNUM_MAX 10
 typedef struct  timerlist_stru
 {
-	char	ref[REFNUM_MAX + 1];
+	char	*ref;
         Timeval time;
 	int	(*callback) (void *);
 	void *	callback_data;
@@ -240,8 +285,10 @@ typedef struct  timerlist_stru
         struct  timerlist_stru *next;
 	long	events;
 	Timeval	interval;
-	int	server;
-	int	window;
+	int	domain;
+	int	domref;
+	int	cancelable;
+	long	fires;
 }       Timer;
 
 static 	Timer *		PendingTimers;
@@ -254,7 +301,7 @@ static Timer * new_timer (void)
 	Timer *	ntimer;
 
 	ntimer = (Timer *) new_malloc(sizeof(Timer));
-	ntimer->ref[0] = 0;
+	ntimer->ref = NULL;
 	ntimer->time.tv_sec = 0;
 	ntimer->time.tv_usec = 0;
 	ntimer->callback = NULL;
@@ -266,8 +313,10 @@ static Timer * new_timer (void)
 	ntimer->events = 0;
 	ntimer->interval.tv_sec = 0;
 	ntimer->interval.tv_usec = 0;
-	ntimer->server = FROMSERV;
-	ntimer->window = -1;
+	ntimer->domain = GENERAL_TIMER;
+	ntimer->domain = -1;
+	ntimer->cancelable = 0;
+	ntimer->fires = 0;
 	return ntimer;
 }
 
@@ -278,7 +327,7 @@ static Timer *clone_timer (Timer *otimer)
 {
 	Timer *ntimer = new_timer();
 
-	strlcpy(ntimer->ref, otimer->ref, sizeof ntimer->ref);
+	malloc_strcpy(&ntimer->ref, otimer->ref);
 	ntimer->time = otimer->time;
 	if ((ntimer->callback = otimer->callback))
 		ntimer->callback_data = otimer->callback_data;
@@ -289,8 +338,10 @@ static Timer *clone_timer (Timer *otimer)
 	ntimer->next = NULL;
 	ntimer->events = otimer->events;
 	ntimer->interval = otimer->interval;
-	ntimer->server = otimer->server;
-	ntimer->window = otimer->window;
+	ntimer->domain = otimer->domain;
+	ntimer->domref = otimer->domref;
+	ntimer->cancelable = otimer->cancelable;
+	ntimer->fires = otimer->fires;
 	return ntimer;
 }
 
@@ -304,6 +355,7 @@ static void delete_timer (Timer *otimer)
 		new_free((char **)&otimer->command);
 		new_free((char **)&otimer->subargs);
 	}
+	new_free(&otimer->ref);
 	new_free((char **)&otimer);
 }
 
@@ -311,6 +363,8 @@ static int	schedule_timer (Timer *ntimer)
 {
 	Timer *tmp, *prev;
 
+	ntimer->fires = 0;
+
 	/* we've created it, now put it in order */
 	for (tmp = PendingTimers; tmp; tmp = tmp->next)
 	{
@@ -387,6 +441,40 @@ int 	timer_exists (const char *ref)
 		return 0;
 }
 
+/*
+ * dump_timers: show all timers in case of emergency
+ */
+void    dump_timers (void)
+{
+        Timer   *tmp;
+        Timeval current;
+        double  time_left;
+
+        yell("*X*X*X*X*X*X*X*X*X* WARNING *X*X*X*X*X*X*X*X*X*X");
+        yell("POLLING LOOP DETECTED -- IMPORTANT DEBUGGING INFO");
+        yell("MAKE SURE TO SAVE THIS VERY IMPORTANT INFORMATION!");
+        yell("%s", empty_string);
+        say("Timer     Seconds   Events Command");
+
+        get_time(&current);
+        for (tmp = PendingTimers; tmp; tmp = tmp->next)
+        {
+                time_left = time_diff(current, tmp->time);
+                if (time_left <= 0)
+                    yell("--> %-10s %-10.2f %-7ld %ld %s", 
+                                tmp->ref, time_left, tmp->events, 
+				tmp->fires,
+                                tmp->callback ? "SYSTEM" : tmp->command);
+                else
+                    yell("    %-10s %-10.2f %-7ld %ld %s", 
+                                tmp->ref, time_left, tmp->events,
+				tmp->fires,
+                                tmp->callback ? "SYSTEM" : tmp->command);
+        }
+        yell("Make sure to give this list to hop on #epic on efnet!");
+        yell("*X*X*X*X*X*X*X*X*X* WARNING *X*X*X*X*X*X*X*X*X*X");
+}
+
 
 /*
  * list_timers:  Display a list of all the TIMER commands that are
@@ -431,16 +519,13 @@ static	void	list_timers (const char *command)
  *
  * "refnum_gets" must be REFNUM_MAX + 1 bytes by definition of API.
  */
-static	int	create_timer_ref (const char *refnum_wanted, char *refnum_gets)
+static	int	create_timer_ref (const char *refnum_wanted, char **refnum_gets)
 {
 	Timer	*tmp;
 	int 	refnum = 0;
 	char	*refnum_want;
 
-	/* Max of 10 characters. */
 	refnum_want = LOCAL_COPY(refnum_wanted);
-	if (strlen(refnum_want) > REFNUM_MAX)
-		refnum_want[REFNUM_MAX] = 0;
 
 	/* If the user doesnt care */
 	if (*refnum_want == 0)
@@ -451,7 +536,7 @@ static	int	create_timer_ref (const char *refnum_wanted, char *refnum_gets)
 			if (refnum < my_atol(tmp->ref))
 				refnum = my_atol(tmp->ref);
 		}
-		strlcpy(refnum_gets, ltoa(refnum+1), REFNUM_MAX + 1);
+		malloc_sprintf(refnum_gets, "%d", refnum + 1);
 	}
 	else
 	{
@@ -459,7 +544,7 @@ static	int	create_timer_ref (const char *refnum_wanted, char *refnum_gets)
 		if (get_timer(refnum_want))
 			return -1;		/* Already in use */
 
-		strlcpy(refnum_gets, refnum_want, REFNUM_MAX + 1);
+		malloc_strcpy(refnum_gets, refnum_want);
 	}
 
 	return 0;
@@ -501,7 +586,7 @@ static void 	remove_all_timers (void)
 	}
 }
 
-static	void	remove_window_timers (int winref)
+static	void	remove_timers_by_domref (int domain, int domref)
 {
 	Timer *ref, *next;
 
@@ -510,7 +595,9 @@ static	void	remove_window_timers (int winref)
 		next = ref->next;
 		if (ref->callback)
 			continue;
-		if (ref->window != winref)
+		if (ref->domain != domain)
+			continue;
+		if (ref->domref != domref)
 			continue;
 		unlink_timer(ref);
 		delete_timer(ref);
@@ -541,10 +628,10 @@ static	void	remove_window_timers (int winref)
  *		 know anything of the nature of the argument.
  * subargs:	 should be NULL, its ignored anyhow.
  */
-char *add_timer (int update, const char *refnum_want, double interval, long events, int (callback) (void *), void *commands, const char *subargs, int winref)
+char *add_timer (int update, const char *refnum_want, double interval, long events, int (callback) (void *), void *commands, const char *subargs, TimerDomain domain, int domref, int cancelable)
 {
 	Timer	*ntimer, *otimer = NULL;
-	char	refnum_got[REFNUM_MAX + 1];
+	char *	refnum_got = NULL;
 	Timeval right_now;
 	char *	retval;
 
@@ -592,26 +679,24 @@ char *add_timer (int update, const char *refnum_want, double interval, long even
 		}
 		else
 		{
-			if (ntimer->callback)
-				ntimer->callback = NULL;
+			ntimer->callback = NULL;
 			malloc_strcpy(&ntimer->command, (const char *)commands);
 			malloc_strcpy(&ntimer->subargs, subargs);
 		}
 
-		if (winref != -1)
-			ntimer->window = winref;
-
+		if (domref != -1)
+			ntimer->domref = domref;
 	}
 	else
 	{
-		if (create_timer_ref(refnum_want, refnum_got) == -1)
+		if (create_timer_ref(refnum_want, &refnum_got) == -1)
 		{
 			say("TIMER: Refnum '%s' already exists", refnum_want);
 			return NULL;
 		}
 
 		ntimer = new_timer();
-		strlcpy(ntimer->ref, refnum_got, sizeof ntimer->ref);
+		ntimer->ref = refnum_got;
 		ntimer->interval = double_to_timeval(interval);
 		ntimer->time = time_add(right_now, ntimer->interval);
 		ntimer->events = events;
@@ -620,10 +705,14 @@ char *add_timer (int update, const char *refnum_want, double interval, long even
 		if (callback)
 			ntimer->callback_data = commands;
 		else
+		{
 			malloc_strcpy(&ntimer->command, (const char *)commands);
-		malloc_strcpy(&ntimer->subargs, subargs);
-		ntimer->window = winref;
-		ntimer->server = from_server;
+			malloc_strcpy(&ntimer->subargs, subargs);
+		}
+		ntimer->domain = domain;
+		ntimer->domref = domref;
+		ntimer->cancelable = cancelable;
+		ntimer->fires = 0;
 	}
 
 	schedule_timer(ntimer);
@@ -632,7 +721,6 @@ char *add_timer (int update, const char *refnum_want, double interval, long even
 }
 
 
-
 /*
  * TimerTimeout:  Called from irc_io to help create the timeout
  * part of the call to select.
@@ -650,6 +738,7 @@ Timeval	TimerTimeout (void)
 
 	get_time(&current);
 	timeout_in = time_subtract(current, PendingTimers->time);
+	PendingTimers->fires++;
 	if (time_diff(right_away, timeout_in) < 0)
 		timeout_in = right_away;
 	return timeout_in;
@@ -688,36 +777,67 @@ void 	ExecuteTimers (void)
 			schedule_timer(next);
 		}
 
-		/*
-		 * Restore from_server and current_window from when the
-		 * timer was registered
-		 */
-		make_window_current_by_refnum(current->window);
-
-		if (is_server_open(current->server))
-			from_server = current->server;
-		else if (is_server_open(current_window->server))
+		if (current->domain == SERVER_TIMER)
+		{
+		    if (!is_server_valid(current->domref))
+		    {
+			if (current->cancelable)
+			    goto advance;
+			/* Otherwise, pretend you were a  "GENERAL" type */
+		    }
+		    else
+		    {
+			from_server = current->domref;
+			make_window_current_by_refnum(
+					get_winref_by_servref(from_server));
+		    }
+		}
+		else if (current->domain == WINDOW_TIMER)
+		{
+		    if (!get_window_by_refnum(current->domref))
+		    {
+			if (current->cancelable)
+			    goto advance;
+			/* Otherwise, pretend you were a "GENERAL" type */
+		    }
+		    else
+		    {
+			make_window_current_by_refnum(current->domref);
 			from_server = current_window->server;
+		    }
+		}
 		else
-			from_server = NOSERV;
+		{
+		    /* General timers focus on the current window. */
+		    if (current_window)
+		    {
+			if (current_window->server != from_server)
+			    from_server = current_window->server;
+		    }
+		    else
+		    {
+			if (from_server != NOSERV)
+			    make_window_current_by_refnum(
+				get_winref_by_servref(from_server));
+		    }
+		}
 
 		/* 
 		 * If a callback function was registered, then
 		 * we use it.  If no callback function was registered,
-		 * then we use ''parse_line''.
+		 * then we call the lambda function.
 		 */
 		get_time(&right_now);
 		now = right_now;
 		if (current->callback)
 			(*current->callback)(current->callback_data);
 		else
-			parse_line("TIMER", current->command, 
-						current->subargs ? 
-						  current->subargs : 
-						  empty_string, 0,0);
+			call_lambda_command("TIMER", current->command,
+							current->subargs);
 
 		from_server = old_from_server;
 		make_window_current_by_refnum(old_refnum);
+advance:
 		delete_timer(current);
 	}
 }
@@ -748,10 +868,10 @@ char *	timerctl (char *input)
 	Timer *	t;
 	int	len;
 
-	GET_STR_ARG(listc, input);
+	GET_FUNC_ARG(listc, input);
 	len = strlen(listc);
 	if (!my_strnicmp(listc, "REFNUM", len)) {
-		GET_STR_ARG(refstr, input);
+		GET_FUNC_ARG(refstr, input);
 		if (!(t = get_timer(refstr)))
 			RETURN_EMPTY;
 		RETURN_STR(t->ref);
@@ -760,23 +880,23 @@ char *	timerctl (char *input)
 		size_t	clue = 0;
 
 		for (t = PendingTimers; t; t = t->next)
-			malloc_strcat_word_c(&retval, space, t->ref, &clue);
+			malloc_strcat_word_c(&retval, space, t->ref, DWORD_DWORDS, &clue);
 		RETURN_MSTR(retval);
 	} else if (!my_strnicmp(listc, "ADD", len)) {
 		RETURN_EMPTY;		/* XXX - Not implemented yet. */
 	} else if (!my_strnicmp(listc, "DELETE", len)) {
-		GET_STR_ARG(refstr, input);
+		GET_FUNC_ARG(refstr, input);
 		if (!(t = get_timer(refstr)))
 			RETURN_EMPTY;
 		if (t->callback)
 			RETURN_EMPTY;
 		RETURN_INT(remove_timer(refstr));
 	} else if (!my_strnicmp(listc, "GET", len)) {
-		GET_STR_ARG(refstr, input);
+		GET_FUNC_ARG(refstr, input);
 		if (!(t = get_timer(refstr)))
 			RETURN_EMPTY;
 
-		GET_STR_ARG(listc, input);
+		GET_FUNC_ARG(listc, input);
 		len = strlen(listc);
 		if (!my_strnicmp(listc, "TIMEOUT", len)) {
 			return malloc_sprintf(NULL, "%ld %ld", (long) t->time.tv_sec,
@@ -795,12 +915,16 @@ char *	timerctl (char *input)
 			return malloc_sprintf(NULL, "%ld %ld", (long) t->interval.tv_sec,
 						    (long) t->interval.tv_usec);
 		} else if (!my_strnicmp(listc, "SERVER", len)) {
-			RETURN_INT(t->server);
+			if (t->domain != SERVER_TIMER)
+				RETURN_INT(-1);
+			RETURN_INT(t->domref);
 		} else if (!my_strnicmp(listc, "WINDOW", len)) {
-			RETURN_INT(t->window);
+			if (t->domain != WINDOW_TIMER)
+				RETURN_INT(-1);
+			RETURN_INT(t->domref);
 		}
 	} else if (!my_strnicmp(listc, "SET", len)) {
-		GET_STR_ARG(refstr, input);
+		GET_FUNC_ARG(refstr, input);
 		if (!(t = get_timer(refstr)))
 			RETURN_EMPTY;
 
@@ -808,7 +932,7 @@ char *	timerctl (char *input)
 		if (t->callback)
 			RETURN_EMPTY;
 
-		GET_STR_ARG(listc, input);
+		GET_FUNC_ARG(listc, input);
 		len = strlen(listc);
 		if (!my_strnicmp(listc, "TIMEOUT", len)) {
 			time_t	tv_sec;
@@ -839,12 +963,14 @@ char *	timerctl (char *input)
 			int	refnum;
 
 			GET_INT_ARG(refnum, input);
-			t->server = refnum;
+			t->domain = SERVER_TIMER;
+			t->domref = refnum;
 		} else if (!my_strnicmp(listc, "WINDOW", len)) {
-			int	winref;
+			int	refnum;
 
-			GET_INT_ARG(winref, input);
-			t->window = winref;
+			GET_INT_ARG(refnum, input);
+			t->domain = WINDOW_TIMER;
+			t->domref = refnum;
 		}
 	} else
 		RETURN_EMPTY;
@@ -852,3 +978,23 @@ char *	timerctl (char *input)
 	RETURN_EMPTY;
 }
 
+/*
+ * The /WINDOW NUMBER command actually swaps the refnums of two windows:
+ * It's possible that 'newref' isn't in use, so that's ok.
+ */
+void    timers_swap_winrefs (int oldref, int newref)
+{
+	Timer *ref;
+
+	for (ref = PendingTimers; ref; ref = ref->next)
+        {
+                if (ref->domain != WINDOW_TIMER)
+                        continue;
+
+		if (ref->domref == newref)
+			ref->domref = oldref;
+		else if (ref->domref == oldref)
+			ref->domref = newref;
+        }
+}
+
