@@ -1,4 +1,4 @@
-/* $EPIC: words.c,v 1.20 2004/11/10 03:20:36 jnelson Exp $ */
+/* $EPIC: words.c,v 1.23 2007/04/12 03:24:14 jnelson Exp $ */
 /*
  * words.c -- right now it just holds the stuff i wrote to replace
  * that beastie arg_number().  Eventually, i may move all of the
@@ -42,15 +42,24 @@
 /* 
  * If "extended" is 1, then honor /xdebug extractw.
  * If "extended" is other than 1 but not zero, imply /xdebug extractw.
+ *
+ * This should turn to YES if DWORD_YES and DEBUG_DWORD ...
  */
 #define CHECK_EXTENDED_SUPPORT					\
-	if (extended == DWORD_YES && 				\
-			((x_debug & DEBUG_EXTRACTW) == 0))	\
-		extended = DWORD_NEVER;				\
-        else if (extended == DWORD_ALWAYS)			\
-		extended = DWORD_YES;				\
-
-
+	if (extended == DWORD_EXTRACTW)				\
+	{							\
+		if ((x_debug & DEBUG_EXTRACTW) == 0)		\
+			extended = DWORD_NO;			\
+		else						\
+			extended = DWORD_YES;			\
+	}							\
+	else if (extended == DWORD_DWORDS)			\
+	{							\
+		if ((x_debug & DEBUG_DWORD) == 0)		\
+			extended = DWORD_NO;			\
+		else						\
+			extended = DWORD_YES;			\
+	}
 
 /*
  * search() looks for a character forward or backward from mark 
@@ -94,7 +103,7 @@ static const char *	find_forward_character (const char *input, const char *start
 	else
 	{
 		simple = 0;
-		what = -1;
+		what = 255;
 	}
 
 	/*
@@ -164,7 +173,7 @@ static const char *	find_backward_quote (const char *input, const char *start, c
 	else
 	{
 		simple = 0;
-		what = -1;
+		what = 255;
 	}
 
 	/*
@@ -287,7 +296,7 @@ static int	move_to_prev_word (const char **str, const char *start, int extended,
 		if (x_debug & DEBUG_EXTRACTW_DEBUG)
 			yell(".... move_to_prev_word: Expensive processing");
 		simple = 0;
-		what = -1;
+		what = 255;
 	}
 
 	/* Overhead -- work out if we're doing "" support or not. */
@@ -333,7 +342,7 @@ static int	move_to_prev_word (const char **str, const char *start, int extended,
 			yell(".... move_to_prev_word: Handling extended word.");
 
 		if (!(before = find_backward_quote(pos, start, delims)))
-			panic("find_backward returned NULL [2]");
+			panic(1, "find_backward returned NULL [2]");
 
 		if (x_debug & DEBUG_EXTRACTW_DEBUG)
 			yell(".... move_to_prev_word: Extended word begins at [%s] (of [%s])", before, start);
@@ -448,9 +457,6 @@ static int	move_to_next_word (const char **str, const char *start, int extended,
 {
 	char	what;
 	int	simple;
-#if 0
-	const char *	mark2;
-#endif
 	const char *	pos;
 
 	/*
@@ -481,7 +487,7 @@ static int	move_to_next_word (const char **str, const char *start, int extended,
 		if (x_debug & DEBUG_EXTRACTW_DEBUG)
 			yell(".... move_to_next_word: Expensive processing");
 		simple = 0;
-		what = -1;
+		what = 255;
 	}
 
 	/*
@@ -527,7 +533,7 @@ static int	move_to_next_word (const char **str, const char *start, int extended,
 		    yell(".... move_to_next_word: handling extended word");
 
 		if (!(after = find_forward_character(pos, start, delims)))
-			panic("find_forward returned NULL [1]");
+			panic(1, "find_forward returned NULL [1]");
 		if (*after)
 			after++;
 		pos = after;
@@ -550,38 +556,8 @@ static int	move_to_next_word (const char **str, const char *start, int extended,
 		    yell(".... move_to_next_word: at next space: [%s]", pos);
 	}
 
-#if 0
-	/*
-	 * 'pos' now points at the space just after the end of the word;
-	 * This space belongs to us, but any further spaces do not long
-	 * belong to us, unless there are no more words after us!  So we
-	 * need to check this.
-	 */
-	/* -- Is the rest of a string just spaces?  -- */
-	if (x_debug & DEBUG_EXTRACTW_DEBUG)
-	    yell("... move_to_next_word: looking for another word...");
-
-	mark2 = pos;
-	while (*mark2 && my_isspace(*mark2))
-		mark2++;
-
-	if (!*mark2)		/* Only spaces after this.  Ok. */
-	{
-		if (x_debug & DEBUG_EXTRACTW_DEBUG)
-		    yell("... move_to_next_word: didn't find one.");
-		pos = mark2;
-	}
-	/* The start of the next word is after our space. */
-	else
-	{
-		if (x_debug & DEBUG_EXTRACTW_DEBUG)
-		    yell("... move_to_next_word: there's another word.");
-		pos++;
-	}
-#else
 	if (*pos)
 		pos++;
-#endif
 
 	if (x_debug & DEBUG_EXTRACTW_DEBUG)
 	    yell("... move_to_next_word: next word starts with [%s]", pos);
